"""
RasgoQL Custom Errors
"""


class RasgoQLWarning(Warning):
    """
    Base warning attributable to RasgoQL code
    """


class RasgoQLException(Exception):
    """
    Base error attributable to RasgoQL code
    """


class DWCredentialsWarning(RasgoQLException):
    """
    Missing DataWarehouse credentials
    """


class DWConnectionError(RasgoQLException):
    """
    Error from DataWarehouse connection activity
    """


class DWQueryError(RasgoQLException):
    """
    Error from DataWarehouse query activity
    """


class ParameterException(RasgoQLException):
    """
    Problem populating a value-restricted parameter
    """


class PackageDependencyWarning(RasgoQLException):
    """
    Problem importing a required package dependency
    """


class SQLWarning(RasgoQLException):
    """
    Possible problem with SQL text
    """


class TableAccessError(RasgoQLException):
    """
    Error accessing Table
    """


class TableConflictException(RasgoQLException):
    """
    Possiblle conflict with Table names
    """


class TransformRenderingError(RasgoQLException):
    """
    Error from Transform rendering activity
    """



"""
Base DataWarehouse classes
"""
from __future__ import annotations
from abc import ABC
import re
import os
from typing import List, Optional, Tuple, Union
from collections import namedtuple

import pandas as pd

FQTN = namedtuple("FQTN", ["database", "schema", "table"], defaults=(None, None, None))


class DWCredentials(ABC):
    """
    Base DW Credentials
    """

    dw_type = None

    @classmethod
    def from_dict(
        cls,
        source_dict: dict,
    ) -> DWCredentials:
        """
        Creates an instance of this Class from a dict
        """
        raise NotImplementedError()

    @classmethod
    def from_env(
        cls,
        filepath: str = None,
    ) -> DWCredentials:
        """
        Creates an instance of this Class from a .env file on your machine
        """
        raise NotImplementedError()

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        raise NotImplementedError()

    def to_env(
        self,
        filepath: str,
        overwrite: bool,
    ):
        """
        Saves credentials to a .env file on your machine
        """
        raise NotImplementedError()

    @staticmethod
    def _parse_env_vars(prefix: str) -> dict[str, Union[str, bool, int]]:
        """
        Pull out all environment variables starting with a prefix
        ( usual same name as data warehouse, i.e. REDSHIFT )
        Auto-detect and convert integers and boolean values.
        """
        prefix = prefix.upper()
        if not prefix.endswith("_"):
            prefix = f"{prefix}_"

        env_vars = {}
        for var_name, value in os.environ.items():
            if var_name.upper().startswith(prefix):
                # Convert Booleans
                if value.lower().strip() in ("false", "true"):
                    value = False if value.lower().strip() == "false" else True
                # Convert Integers
                elif value.strip().isnumeric():
                    value = int(value.strip())
                env_vars[var_name.lower()[len(prefix) :]] = value
        return env_vars


class DataWarehouse(ABC):
    """
    Base DW Class
    """

    dw_type = None
    credentials_class = DWCredentials

    def __init__(self):
        self.credentials = None
        self.connection = None

    # ---------------------------
    # FQTN and namespace methods
    # ---------------------------
    def magic_fqtn_handler(
        self,
        possible_fqtn: str,
        default_namespace: str,
    ) -> str:
        """
        Makes all of your wildest dreams come true... well not *that* one
        """
        input_db, input_schema, table = self.parse_fqtn(possible_fqtn, default_namespace, False)
        default_database, default_schema = self.parse_namespace(default_namespace)
        database = input_db or default_database
        schema = input_schema or default_schema
        return self.make_fqtn(database, schema, table)

    def make_fqtn(
        self,
        database: str,
        schema: str,
        table: str,
    ) -> str:
        """
        Accepts component parts and returns a fully qualified table string
        """
        return f"{database}.{schema}.{table}"

    def make_namespace_from_fqtn(
        self,
        fqtn: str,
    ) -> str:
        """
        Accepts component parts and returns a fully qualified namespace string
        """
        database, schema, _ = self.parse_fqtn(fqtn)
        return f"{database}.{schema}"

    def parse_fqtn(
        self,
        fqtn: str,
        default_namespace: Optional[str] = None,
        strict: bool = True,
    ) -> FQTN:
        """
        Accepts a possible fully qualified table string and returns its component parts
        """
        if strict:
            fqtn = self.validate_fqtn(fqtn)
            return FQTN(*fqtn.split("."))

        if fqtn.count(".") == 2:
            return FQTN(*fqtn.split("."))

        if not default_namespace:
            raise ValueError(f'{fqtn} is not a well-formed fqtn')

        database, schema = self.parse_namespace(default_namespace)

        if fqtn.count(".") == 1:
            return FQTN(database, *fqtn.split("."))
        if fqtn.count(".") == 0:
            return FQTN(database, schema, fqtn)
        raise ValueError(f'{fqtn} is not a well-formed fqtn')

    def parse_namespace(
        self,
        namespace: str,
    ) -> tuple:
        """
        Accepts a possible namespace string and returns its component parts
        """
        namespace = self.validate_namespace(namespace)
        return tuple(namespace.split("."))

    def validate_fqtn(
        self,
        fqtn: str,
    ) -> str:
        """
        Accepts a possible fully qualified table string and decides whether it is well-formed
        """
        if re.match(r'^[^\s]+\.[^\s]+\.[^\s]+', fqtn):
            return fqtn
        raise ValueError(f'{fqtn} is not a well-formed fqtn')

    def validate_namespace(
        self,
        namespace: str,
    ) -> str:
        """
        Accepts a possible namespace string and decides whether it is well-formed
        """
        if re.match(r'^[^\s]+\.[^\s]+', namespace):
            return namespace
        raise ValueError(f'{namespace} is not a well-formed namespace')

    # Core methods

    def change_namespace(
        self,
        namespace: str,
    ):
        """
        Change the default namespace of this connection
        """
        raise NotImplementedError()

    def connect(
        self,
        credentials: Union[dict, DWCredentials],
    ):
        """
        Connect to this DataWarehouse
        """
        raise NotImplementedError()

    def close_connection(self):
        """
        Close connection to this DataWarehouse
        """
        raise NotImplementedError()

    def create(
        self,
        sql: str,
        fqtn: str,
        table_type: str = 'VIEW',
        overwrite: bool = False,
    ):
        """
        Create a view or table from given SQL
        """
        raise NotImplementedError()

    @property
    def default_namespace(self) -> str:
        """
        Returns the default namespace of this connection
        """
        raise NotImplementedError()

    @default_namespace.setter
    def default_namespace(
        self,
        new_namespace: str,
    ):
        """
        Setter method for the `default_namespace` property
        """
        raise NotImplementedError()

    def execute_query(
        self,
        sql: str,
        response: str = 'tuple',
        acknowledge_risk: bool = False,
        **kwargs,
    ):
        """
        Run a query against this DataWarehouse
        """
        raise NotImplementedError()

    def get_ddl(
        self,
        fqtn: str,
    ) -> str:
        """
        Returns the create statement for a table or view
        """
        raise NotImplementedError()

    def get_object_details(
        self,
        fqtn: str,
    ) -> tuple:
        """
        Return details of a table or view in this DataWarehouse
        """
        raise NotImplementedError()

    def get_schema(
        self,
        fqtn: str,
        create_sql: str = None,
    ) -> List[Tuple[str, str]]:
        """
        Return the schema of a table or view
        """
        raise NotImplementedError()

    def list_tables(
        self,
        database: str = None,
        schema: str = None,
    ):
        """
        List all available tables in this DataWarehouse
        """
        raise NotImplementedError()

    def preview(
        self,
        sql: str,
        limit: int = 10,
    ) -> pd.DataFrame:
        """
        Returns 10 records into a pandas DataFrame
        """
        raise NotImplementedError()

    def save_df(
        self,
        df: pd.DataFrame,
        fqtn: str,
        method: str = None,
    ):
        """
        Creates a table in this DataWarehouse from a pandas Dataframe
        """
        raise NotImplementedError()

    # ---------------------------
    # Core Data Warehouse helpers
    # ---------------------------
    def _table_exists(
        self,
        fqtn: str,
    ) -> bool:
        """
        Check for existence of fqtn in this Data Warehouse and return a boolean
        """
        raise NotImplementedError()

    def _validate_namespace(
        self,
        namespace: str,
    ):
        """
        Checks a namespace string for compliance with this DataWarehouse format
        """
        raise NotImplementedError()



"""
Verbose messaging functions
"""
import logging

rql_run_verbose: bool = False


def verbose_message(
    message: str,
    logger: logging.Logger,
) -> None:
    """
    Logs messages only if verbose flag is set to true
    """
    logger.setLevel(logging.INFO)
    if rql_run_verbose:
        logger.info(message)


def set_verbose(
    value: bool,
) -> None:
    """
    Set global var
    """
    global rql_run_verbose
    rql_run_verbose = value



"""
Redshift DataWarehouse classes
"""
from __future__ import annotations
import logging
import os
from typing import Optional, Union
import json

import pandas as pd

from rasgoql.data.base import DWCredentials
from rasgoql.data.sqlalchemy import SQLAlchemyDataWarehouse
from rasgoql.imports import alchemy_engine, alchemy_session, alchemy_url
from rasgoql.utils.creds import load_env, save_env
from rasgoql.errors import DWCredentialsWarning, TableConflictException
from rasgoql.primitives.enums import check_table_type
from rasgoql.utils.messaging import verbose_message


logging.basicConfig()
logger = logging.getLogger("Redshift DataWarehouse")
logger.setLevel(logging.INFO)


class RedshiftCredentials(DWCredentials):
    """
    Redshift Credentials

    Full list of accepted parameters available at
    https://github.com/aws/amazon-redshift-python-driver#connection-parameters
    """

    dw_type = "redshift"

    def __init__(
        self,
        username: str,
        password: str,
        host: str,
        port: Union[str, int],
        database: str,
        schema: str,
        **kwargs,
    ):
        self.username = username
        self.password = password
        self.host = host
        try:
            self.port = int(port)
        except ValueError:
            raise DWCredentialsWarning(f"Redshift port number must be an integer") from None
        self.database = database
        self.schema = schema
        self.conn_params = kwargs

    def __repr__(self) -> str:
        return json.dumps(
            {
                "username": self.username,
                "host": self.host,
                "port": self.port,
                "database": self.database,
                "schema": self.schema,
                "conn_params": self.conn_params,
            }
        )

    @classmethod
    def from_env(cls, filepath: Optional[os.PathLike] = None) -> RedshiftCredentials:
        """
        Creates an instance of this Class from a .env file on your machine
        """
        load_env(filepath)
        env_vars = cls._parse_env_vars("REDSHIFT_")

        try:
            username = env_vars.pop("user" if "user" in env_vars else "username")
            password = env_vars.pop("password")
            host = env_vars.pop("host")
            port = int(env_vars.pop("port"))
            database = env_vars.pop("database")
            schema = env_vars.pop("schema")
        except KeyError as key_name:
            raise DWCredentialsWarning(
                f"Your env file is missing expected credential variable {key_name}. Consider running "
                "RedshiftCredentials(*args).to_env() to repair this."
            ) from None
        except ValueError:
            raise DWCredentialsWarning(f"Redshift port number must be an integer") from None
        return cls(username, password, host, port, database, schema, **env_vars)

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        return {
            "username": self.username,
            "password": self.password,
            "host": self.host,
            "port": self.port,
            "database": self.database,
            "schema": self.schema,
            "dw_type": self.dw_type,
            "conn_params": self.conn_params,
        }

    def to_env(self, filepath: str = None, overwrite: bool = False):
        """
        Saves credentials to a .env file on your machine
        """
        creds = {
            "REDSHIFT_USERNAME": self.username,
            "REDSHIFT_PASSWORD": self.password,
            "REDSHIFT_HOST": self.host,
            "REDSHIFT_PORT": self.port,
            "REDSHIFT_DATABASE": self.database,
            "REDSHIFT_SCHEMA": self.schema,
            **{f"REDSHIFT_{key}": str(value) for key, value in self.conn_params.items()},
        }
        return save_env(creds, filepath, overwrite)


class RedshiftDataWarehouse(SQLAlchemyDataWarehouse):
    """
    Redshift DataWarehouse
    """

    dw_type = "redshift"
    credentials_class = RedshiftCredentials

    def __init__(self):
        super().__init__()
        self.credentials: dict
        self.connection: alchemy_session
        self.database: Optional[str] = None
        self.schema: Optional[str] = None

    # ---------------------------
    # Core Data Warehouse methods
    # ---------------------------
    def change_namespace(self, namespace: str) -> None:
        """
        Changes the default namespace of your connection
        Params:
        `namespace`: str:
            namespace (database.schema)
        """
        raise NotImplementedError(
            "Connecting to a new Database in a single session is not supported by Redshift. "
            "Please build a new connection using the RedshiftCredentials class"
        )

    def connect(self, credentials: Union[dict, RedshiftCredentials]):
        """
        Connect to Redshift

        Params:
        `credentials`: dict:
            dict (or RedshiftCredentials class) holding the connection credentials
        """
        if isinstance(credentials, RedshiftCredentials):
            credentials = credentials.to_dict()

        try:
            credentials["conn_params"]["application_name"] = "rasgoql"
        except KeyError:
            credentials["conn_params"] = {"application_name": "rasgoql"}

        super().connect(credentials)

    def close_connection(self):
        """
        Close connection to Redshift
        """
        try:
            if self.connection:
                self.connection.close()
            self.connection = None
            verbose_message("Connection to Redshift closed", logger)
        except Exception as e:
            self._error_handler(e)

    def create(self, sql: str, fqtn: str, table_type: str = "VIEW", overwrite: bool = False):
        """
        Create a view or table from given SQL

        Params:
        `sql`: str:
            query that returns data (i.e. can be wrapped in a CREATE TABLE statement)
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
            Name for the new table
        `table_type`: str:
            One of values: [view, table]
        `overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        table_type = check_table_type(table_type)
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        _, schema_name, table_name = self.parse_fqtn(fqtn)
        if self._table_exists(fqtn) and not overwrite:
            msg = (
                f"A table or view named {fqtn} already exists. "
                "If you are sure you want to overwrite it, "
                "pass in overwrite=True and run this function again"
            )
            raise TableConflictException(msg)
        query = f"CREATE OR REPLACE {table_type} {schema_name}.{table_name} AS {sql}"
        self.execute_query(query, acknowledge_risk=True, response="None")
        return fqtn

    @property
    def default_namespace(self) -> str:
        """
        Returns the default database.schema of this connection
        """
        return f"{self.database}.{self.schema}"

    @default_namespace.setter
    def default_namespace(self, new_namespace: str):
        """
        Setter method for the `default_namespace` property
        """
        namespace = self.validate_namespace(new_namespace)
        db, schema = self.parse_namespace(namespace)
        self.database = db
        self.schema = schema

    def get_ddl(self, fqtn: str) -> pd.DataFrame:
        return super().get_ddl(fqtn)

    def get_object_details(self, fqtn: str) -> tuple:
        """
        Return details of a table or view in Redshift
        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        Response:
            object exists: bool
            is rasgo object: bool
            object type: [table|view|unknown]
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        database, schema, table = self.parse_fqtn(fqtn)
        sql = (
            f"SELECT EXISTS(SELECT 1 FROM pg_catalog.pg_class c JOIN "
            f"pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE "
            f"n.nspname = '{schema}' AND c.relname = '{table}')"
        )
        result = self.execute_query(sql, response="dict")
        obj_exists = len(result) > 0
        is_rasgo_obj = False
        obj_type = "unknown"
        return obj_exists, is_rasgo_obj, obj_type

    # --------------------------
    # Redshift specific helpers
    # --------------------------
    @property
    def _engine(self) -> alchemy_engine:
        """
        Returns a SQLAlchemy engine
        """
        url = alchemy_url.create(
            drivername="redshift+redshift_connector",
            database=self.database,
            username=self.credentials.get("username"),
            password=self.credentials.get("password"),
            host=self.credentials.get("host"),
            port=self.credentials.get("port"),
        )
        return alchemy_engine(
            url,
            connect_args=self.credentials.get("conn_params", {}),
        )



"""
Postgres DataWarehouse classes
"""
import logging
import os
from typing import Union
from urllib.parse import quote_plus as urlquote

import json
import pandas as pd

from rasgoql.data.base import DWCredentials
from rasgoql.data.sqlalchemy import SQLAlchemyDataWarehouse

from rasgoql.errors import (
    DWCredentialsWarning,
    PackageDependencyWarning,
    TableConflictException,
)
from rasgoql.imports import alchemy_engine, alchemy_session
from rasgoql.primitives.enums import check_table_type
from rasgoql.utils.creds import load_env, save_env
from rasgoql.utils.messaging import verbose_message

logging.basicConfig()
logger = logging.getLogger("Postgres DataWarehouse")
logger.setLevel(logging.INFO)


class PostgresCredentials(DWCredentials):
    """
    Postgres Credentials
    """

    dw_type = "postgresql"

    def __init__(
        self,
        username: str,
        password: str,
        host: str,
        port: str,
        database: str,
        schema: str,
    ):
        if alchemy_engine is None:
            raise PackageDependencyWarning(
                "Missing a required python package to run Postgres. "
                "Please download the Postgres package by running: "
                "pip install rasgoql[postgres]"
            )
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.database = database
        self.schema = schema

    def __repr__(self) -> str:
        return json.dumps(
            {
                "user": self.username,
                "host": self.host,
                "port": self.port,
                "database": self.database,
                "schema": self.schema,
            }
        )

    @classmethod
    def from_env(cls, filepath: str = None) -> "PostgresCredentials":
        """
        Creates an instance of this Class from a .env file on your machine
        """
        load_env(filepath)
        username = os.getenv("POSTGRES_USERNAME")
        password = os.getenv("POSTGRES_PASSWORD")
        host = os.getenv("POSTGRES_HOST")
        port = os.getenv("POSTGRES_PORT")
        database = os.getenv("POSTGRES_DATABASE")
        schema = os.getenv("POSTGRES_SCHEMA")
        if not all([username, password, host, port, database, schema]):
            raise DWCredentialsWarning(
                "Your env file is missing expected credentials. Consider running "
                "PostgresCredentials(*args).to_env() to repair this."
            )
        return cls(username, password, host, port, database, schema)

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        return {
            "username": self.username,
            "password": self.password,
            "host": self.host,
            "port": self.port,
            "database": self.database,
            "schema": self.schema,
            "dw_type": self.dw_type,
        }

    def to_env(self, filepath: str = None, overwrite: bool = False):
        """
        Saves credentials to a .env file on your machine
        """
        creds = {
            "POSTGRES_USERNAME": self.username,
            "POSTGRES_PASSWORD": self.password,
            "POSTGRES_HOST": self.host,
            "POSTGRES_PORT": self.port,
            "POSTGRES_DATABASE": self.database,
            "POSTGRES_SCHEMA": self.schema,
        }
        return save_env(creds, filepath, overwrite)


class PostgresDataWarehouse(SQLAlchemyDataWarehouse):
    """
    Postgres DataWarehouse
    """

    dw_type = "postgresql"
    credentials_class = PostgresCredentials

    def __init__(self):
        super().__init__()

    # ---------------------------
    # FQTN and namespace methods
    # ---------------------------
    def parse_table_and_schema_from_fqtn(
        self,
        fqtn: str,
    ) -> tuple:
        """
        Accepts a possible FQTN and returns the schema and table from it
        """
        fqtn = self.validate_fqtn(fqtn)
        return tuple(fqtn.split(".")[1:])

    # ---------------------------
    # Core Data Warehouse methods
    # ---------------------------
    def change_namespace(self, namespace: str) -> None:
        """
        Changes the default namespace of your connection
        Params:
        `namespace`: str:
            namespace (database.schema)
        """
        raise NotImplementedError(
            "Connecting to a new Database in a single session is not supported by Postgres. "
            "Please build a new connection using the PostgresCredentials class"
        )

    def connect(self, credentials: Union[dict, PostgresCredentials]):
        """
        Connect to Postgres
        Params:
        `credentials`: dict:
            dict (or PostgresCredentials class) holding the connection credentials
        """
        super().connect(credentials)

    def create(
        self,
        sql: str,
        fqtn: str,
        table_type: str = "VIEW",
        overwrite: bool = False,
    ):
        """
        Create a view or table from given SQL
        Params:
        `sql`: str:
            query that returns data (i.e. can be wrapped in a CREATE TABLE statement)
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
            Name for the new table
        `table_type`: str:
            One of values: [view, table]
        `overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        table_type = check_table_type(table_type)
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        schema, table = self.parse_table_and_schema_from_fqtn(fqtn=fqtn)
        if self._table_exists(fqtn=fqtn) and not overwrite:
            msg = (
                f"A table or view named {fqtn} already exists. "
                "If you are sure you want to overwrite it, "
                "pass in overwrite=True and run this function again"
            )
            raise TableConflictException(msg)
        query = f"CREATE OR REPLACE {table_type} {schema}.{table} AS {sql}"
        self.execute_query(query, acknowledge_risk=True, response="None")
        return fqtn

    def get_ddl(self, fqtn: str) -> pd.DataFrame:
        """
        Returns a DataFrame describing the column in the table
        `fqtn`: str:
            Fully-qualified Table Name (database.schema.table)
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        _, schema_name, table_name = self.parse_fqtn(fqtn)
        sql = (
            f"select table_schema, table_name, column_name, data_type, "
            f"character_maximum_length, column_default, is_nullable from "
            f"INFORMATION_SCHEMA.COLUMNS where table_name = '{table_name}' "
            f"and table_schema = '{schema_name}';"
        )
        query_response = self.execute_query(sql, response="DF")
        return query_response

    def get_object_details(self, fqtn: str) -> tuple:
        """
        Return details of a table or view in Postgres
        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        Response:
            object exists: bool
            is rasgo object: bool
            object type: [table|view|unknown]
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        database, schema, table = self.parse_fqtn(fqtn)
        sql = (
            f"SELECT EXISTS(SELECT 1 FROM pg_catalog.pg_class c JOIN "
            f"pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE "
            f"n.nspname = '{schema}' AND    c.relname = '{table}')"
        )
        result = self.execute_query(sql, response="dict")
        obj_exists = len(result) > 0
        is_rasgo_obj = False
        obj_type = "unknown"
        return obj_exists, is_rasgo_obj, obj_type

    # --------------------------
    # Postgres specific helpers
    # --------------------------
    @property
    def _engine(self) -> "alchemy_engine":
        """
        Returns a SQLAlchemy engine
        """
        engine_url = (
            f"{self.credentials.get('dw_type')}://"
            f"{self.credentials.get('username')}:"
            f"{urlquote(self.credentials.get('password'))}"
            f"@{self.credentials.get('host')}:"
            f"{self.credentials.get('port')}/"
            f"{self.credentials.get('database')}"
        )
        return alchemy_engine(engine_url)



"""
MySQL DataWarehouse classes
"""
import logging
import os
import re
from typing import Union
from urllib.parse import quote_plus as urlquote

import json
import pandas as pd

from rasgoql.data.base import DWCredentials
from rasgoql.data.sqlalchemy import SQLAlchemyDataWarehouse

from rasgoql.errors import (
    DWCredentialsWarning,
    PackageDependencyWarning,
    TableConflictException,
    ParameterException,
)
from rasgoql.imports import alchemy_engine, alchemy_session
from rasgoql.primitives.enums import check_table_type
from rasgoql.utils.creds import load_env, save_env
from rasgoql.utils.messaging import verbose_message

logging.basicConfig()
logger = logging.getLogger("MySQL DataWarehouse")
logger.setLevel(logging.INFO)


class MySQLCredentials(DWCredentials):
    """
    MySQL Credentials
    """

    dw_type = "mysql"
    db_api = "pymysql"

    def __init__(self, username: str, password: str, host: str, database: str):
        if alchemy_engine is None:
            raise PackageDependencyWarning(
                "Missing a required python package to run MySQL. "
                "Please download the MySQL package by running: "
                "pip install rasgoql[mysql]"
            )
        self.username = username
        self.password = password
        self.host = host
        self.database = database

    def __repr__(self) -> str:
        return json.dumps(
            {
                "user": self.username,
                "host": self.host,
                "database": self.database,
            }
        )

    @classmethod
    def from_env(cls, filepath: str = None) -> "MySQLCredentials":
        """
        Creates an instance of this Class from a .env file on your machine
        """
        load_env(filepath)
        username = os.getenv("MYSQL_USERNAME")
        password = os.getenv("MYSQL_PASSWORD")
        host = os.getenv("MYSQL_HOST")
        database = os.getenv("MYSQL_DATABASE")
        if not all([username, password, host, database]):
            raise DWCredentialsWarning(
                "Your env file is missing expected credentials. Consider running "
                "MySQLCredentials(*args).to_env() to repair this."
            )
        return cls(username, password, host, database)

    def to_dict(self) -> dict:
        """
        Returns a dict of the credentials
        """
        return {
            "username": self.username,
            "password": self.password,
            "host": self.host,
            "database": self.database,
            "dw_type": self.dw_type,
            "db_api": self.db_api,
        }

    def to_env(self, filepath: str = None, overwrite: bool = False):
        """
        Saves credentials to a .env file on your machine
        """
        creds = {
            "MYSQL_USERNAME": self.username,
            "MYSQL_PASSWORD": self.password,
            "MYSQL_HOST": self.host,
            "MYSQL_DATABASE": self.database,
        }
        return save_env(creds, filepath, overwrite)


class MySQLDataWarehouse(SQLAlchemyDataWarehouse):
    """
    MySQL DataWarehouse
    """

    dw_type = "mysql"
    credentials_class = MySQLCredentials

    def __init__(self):
        super().__init__()

    # ---------------------------
    # FQTN and namespace methods
    # ---------------------------
    # FQTNs in MySQL DBs consist only of "DB"."TABLE". This means that a MySQL
    # namespace should be only the DB name
    def magic_fqtn_handler(
        self,
        possible_fqtn: str,
        default_namespace: str,
    ) -> str:
        """
        Makes all of your wildest dreams come true... well not *that* one
        """
        input_db, table = self.parse_fqtn(possible_fqtn, default_namespace, False)
        default_database = self.parse_namespace(default_namespace)
        database = input_db or default_database
        return self.make_fqtn(database, table)

    def make_fqtn(
        self,
        database: str,
        table: str,
    ) -> str:
        """
        Accepts component parts and returns a fully qualified table string
        """
        return f"{database}.{table}"

    def make_namespace_from_fqtn(
        self,
        fqtn: str,
    ) -> str:
        """
        Accepts component parts and returns a fully qualified namespace string
        """
        database, _ = self.parse_fqtn(fqtn)
        return f"{database}"

    def parse_fqtn(
        self,
        fqtn: str,
        default_namespace: str = None,
        strict: bool = True,
    ) -> tuple:
        """
        Accepts a possible fully qualified table string and returns its component parts
        """
        if strict:
            fqtn = self.validate_fqtn(fqtn)
            return (*fqtn.split("."),)
        database = self.parse_namespace(default_namespace)
        if fqtn.count(".") == 1:
            return (*fqtn.split("."),)
        if fqtn.count(".") == 0:
            return (database, fqtn)
        raise ValueError(f'{fqtn} is not a well-formed fqtn')

    def parse_namespace(
        self,
        namespace: str,
    ) -> str:
        """
        Accepts a possible namespace string and returns its component parts
        """
        namespace = self.validate_namespace(namespace)
        return namespace

    def validate_fqtn(self, fqtn: str) -> str:
        """
        Accepts a possible fully qualified table string and decides whether it is well-formed
        """
        if re.match(r'^[^\s]+\.[^\s]+', fqtn):
            return fqtn
        raise ValueError(f'{fqtn} is not a well-formed fqtn')

    def validate_namespace(
        self,
        namespace: str,
    ) -> str:
        """
        Accepts a possible namespace string and decides whether it is well-formed
        """
        if namespace.count(".") == 0:
            return namespace
        raise ValueError(f'{namespace} is not a well-formed namespace')

    # ---------------------------
    # Core Data Warehouse methods
    # ---------------------------

    @property
    def default_namespace(self) -> str:
        """
        Returns the default database of this connection
        """
        return f"{self.database}"

    @default_namespace.setter
    def default_namespace(self, new_namespace: str):
        self.database = new_namespace

    def change_namespace(self, namespace: str) -> None:
        """
        Changes the default namespace of your connection

        Params:
        `namespace`: str:
            namespace (database.schema)
        """
        raise NotImplementedError(
            "Connecting to a new Database in a single session is not supported by MySQL. "
            "Please build a new connection using the MySQLCredentials class"
        )

    def connect(self, credentials: Union[dict, MySQLCredentials]):
        """
        Connect to MySQL

        Params:
        `credentials`: dict:
            dict (or MySQLCredentials class) holding the connection credentials
        """
        super().connect(credentials)

    def create(
        self,
        sql: str,
        fqtn: str,
        table_type: str = "VIEW",
        overwrite: bool = False,
    ):
        """
        Create a view or table from given SQL

        Params:
        `sql`: str:
            query that returns data (i.e. can be wrapped in a CREATE TABLE statement)
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
            Name for the new table
        `table_type`: str:
            One of values: [view, table]
        `overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        table_type = check_table_type(table_type)
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        if self._table_exists(fqtn=fqtn) and not overwrite:
            msg = (
                f"A table or view named {fqtn} already exists. "
                "If you are sure you want to overwrite it, "
                "pass in overwrite=True and run this function again"
            )
            raise TableConflictException(msg)
        query = f"CREATE OR REPLACE {table_type} {fqtn} AS {sql}"
        self.execute_query(query, acknowledge_risk=True, response="None")
        return fqtn

    def get_ddl(self, fqtn: str) -> pd.DataFrame:
        """
        Returns a DataFrame describing the column in the table

        `fqtn`: str:
            Fully-qualified Table Name (database.schema.table)
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        db, table_name = self.parse_fqtn(fqtn, default_namespace=self.default_namespace, strict=False)
        sql = f"SHOW CREATE TABLE {db}.{table_name}"
        query_response = self.execute_query(sql, response="DF")
        return query_response

    def get_object_details(self, fqtn: str) -> tuple:
        """
        Return details of a table or view in MySQL

        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)

        Response:
            object exists: bool
            is rasgo object: bool
            object type: [table|view|unknown]
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        db, table = self.parse_fqtn(fqtn, default_namespace=self.default_namespace, strict=False)
        sql = f"SHOW TABLES IN {db} LIKE '{table}'"
        result = self.execute_query(sql, response="dict")
        obj_exists = len(result) > 0
        is_rasgo_obj = False
        obj_type = "unknown"
        return obj_exists, is_rasgo_obj, obj_type

    # TODO: delete unused code?
    # def preview(self, sql: str = None, limit: int = 10) -> pd.DataFrame:
    #     """
    #     Returns 10 records into a pandas DataFrame

    #     Params:
    #     `sql`: str:
    #         SQL statment passed from calling method
    #         This is normally set in the preview method of a transform or dataset,
    #         but must be overridden for MySQL previews because MySQL does not
    #         utliize standard FQTNs
    #     `limit`: int:
    #         Records to return
    #     """
    #     passed_fqtn = sql.split("FROM ")[1]
    #     db, _, table = parse_fqtn(
    #         passed_fqtn, default_namespace=self.default_namespace, strict=False
    #     )
    #     return self.execute_query(
    #         f"SELECT * FROM {db}.{table} LIMIT {limit}",
    #         response="df",
    #         acknowledge_risk=True,
    #     )

    # --------------------------
    # MySQL specific helpers
    # --------------------------
    @property
    def _engine(self) -> "alchemy_engine":
        """
        Returns a SQLAlchemy engine
        """
        engine_url = (
            f"{self.credentials.get('dw_type')}"
            f"+{self.credentials.get('db_api')}://"
            f"{self.credentials.get('username')}:"
            f"{urlquote(self.credentials.get('password'))}"
            f"@{self.credentials.get('host')}/"
            f"{self.credentials.get('database')}"
        )
        return alchemy_engine(engine_url)



import json
import contextlib
from multiprocessing.pool import ThreadPool
import atexit
from typing import Optional, Any

import requests

HEAP_KEY = "2288014214"
HEAP_URL = "https://heapanalytics.com/api"

thread_pool = ThreadPool(2)
atexit.register(thread_pool.terminate)


def failure_telemetry(
    obj: Any,
    name: str,
):
    attr = object.__getattribute__(obj, name)
    # TODO: change check to isinstance(attr, Callable)
    if hasattr(attr, "__call__") and not name.startswith("_") and not name.startswith("connect"):

        def logged_function(*args, **kwargs):
            tracked_properties = {
                "source": "RasgoQL",
                "class": obj.__class__.__name__,
                "module": attr.__module__,
                "method": attr.__name__,
            }

            try:
                result = attr(*args, **kwargs)
            except Exception as err:
                tracked_properties["execution_status"] = "failed"
                thread_pool.apply_async(track_call, args=(tracked_properties, f"{err.__class__.__name__}: {err}"))
                raise err from None
            else:
                tracked_properties["execution_status"] = "completed"
                thread_pool.apply_async(track_call, args=(tracked_properties, None))
                return result

        return logged_function
    return attr


def track_call(
    tracked_properties: dict,
    error_message: Optional[str],
):
    """
    Send a "track" event to the Heap Analytics API server.
    :param tracked_properties:  contains event data for heap to track
    :param error_message: if the called function fails, this string will
    be the failure message
    """
    data = {
        "app_id": HEAP_KEY,
        # TODO: implement cookies and opt-out for individual level tracking
        "identity": "user",
        "event": tracked_properties["method"],
        "properties": tracked_properties,
    }
    if error_message:
        data["properties"]["error_message"] = error_message

    with contextlib.suppress(Exception):
        requests.post(
            url=f"{HEAP_URL}/track",
            data=json.dumps(data),
            headers={"Content-Type": "application/json"},
            timeout=60,
        )



"""
Generic SQLAlchemy DataWarehouse classes
"""
from __future__ import annotations
from abc import abstractmethod
import logging
from typing import List, Optional, Tuple, Union
from urllib.parse import quote_plus as urlquote

import pandas as pd

from rasgoql.data.base import DataWarehouse, DWCredentials
from rasgoql.errors import (
    DWConnectionError,
    SQLWarning,
    TableAccessError,
    TableConflictException,
)
from rasgoql.imports import alchemy_engine, alchemy_session, alchemy_exceptions
from rasgoql.primitives.enums import (
    check_response_type,
    check_table_type,
    check_write_method,
)
from rasgoql.utils.df import cleanse_sql_dataframe, generate_dataframe_ddl
from rasgoql.utils.messaging import verbose_message
from rasgoql.utils.sql import is_scary_sql

logging.basicConfig()
logger = logging.getLogger("SQLAlchemy DataWarehouse")
logger.setLevel(logging.INFO)

# Each DB derived from the generic SQLAlchemy class will have slightly
# different connection strings. Build a unique Credentials class for each DB


class SQLAlchemyDataWarehouse(DataWarehouse):
    """
    Base SQLAlchemy DataWarehouse
    """

    dw_type = None
    credentials_class = None

    def __init__(self):
        super().__init__()
        self.credentials: Optional[Union[dict, DWCredentials]]
        self.connection: alchemy_session
        self.database = None
        self.schema = None

    # ---------------------------
    # Core Data Warehouse methods
    # ---------------------------
    def change_namespace(self, namespace: str) -> None:
        """
        Changes the default namespace of your connection
        Params:
        `namespace`: str:
            namespace (database.schema)
        """
        raise NotImplementedError(
            "Connecting to a new Database in a single session is not supported "
            "by the SQLAlchemy Engine. Please build a new connection."
        )

    @abstractmethod
    def connect(self, credentials: Union[dict, DWCredentials]):
        """
        Connect to DB
        Params:
        `credentials`: dict:
            dict  holding the connection credentials
        """
        if isinstance(credentials, DWCredentials):
            credentials = credentials.to_dict()

        try:
            self.credentials = credentials
            self.database = credentials.get("database")
            self.schema = credentials.get("schema")
            self.connection = alchemy_session(self._engine, autocommit=True)
        except Exception as e:
            self._error_handler(e)

    def close_connection(self):
        """
        Close connection
        """
        try:
            if self.connection:
                self.connection.close()
            self.connection = None
            verbose_message("Connection closed", logger)
        except Exception as e:
            self._error_handler(e)

    def create(
        self,
        sql: str,
        fqtn: str,
        table_type: str = "VIEW",
        overwrite: bool = False,
    ):
        """
        Create a view or table from given SQL
        If a derived class's DB cannot execute against a FQTN (e.g., Postgres),
        override this method or SQL execution will fail
        Params:
        `sql`: str:
            query that returns data (i.e. can be wrapped in a CREATE TABLE statement)
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
            Name for the new table
        `table_type`: str:
            One of values: [view, table]
        `overwrite`: bool
            pass True when this table name already exists in your DataWarehouse
            and you know you want to overwrite it
            WARNING: This will completely overwrite data in the existing table
        """
        table_type = check_table_type(table_type)
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        if self._table_exists(fqtn=fqtn) and not overwrite:
            msg = (
                f"A table or view named {fqtn} already exists. "
                "If you are sure you want to overwrite it, "
                "pass in overwrite=True and run this function again"
            )
            raise TableConflictException(msg)
        query = f"CREATE OR REPLACE {table_type} {fqtn} AS {sql}"
        self.execute_query(query, acknowledge_risk=True, response="None")
        return fqtn

    @property
    def default_namespace(self) -> str:
        """
        Returns the default database.schema of this connection
        """
        return f"{self.database}.{self.schema}"

    @default_namespace.setter
    def default_namespace(self, new_namespace: str):
        namespace = self._validate_namespace(new_namespace)
        db, schema = self.parse_namespace(namespace)
        self.default_database = db
        self.default_schema = schema

    def execute_query(
        self,
        sql: str,
        response: str = "tuple",
        acknowledge_risk: bool = False,
        **kwargs,
    ) -> Union[list[dict], pd.DataFrame, list[tuple], None]:
        """
        Run a query against DB and return all results
        `sql`: str:
            query text to execute
        `response`: str:
            Possible values: [dict, df, None]
        `acknowledge_risk`: bool:
            pass True when you know your SQL statement contains
            a potentially dangerous or data-altering operation
            and still want to run it against your DataWarehouse
        """
        response = check_response_type(response)
        if is_scary_sql(sql) and not acknowledge_risk:
            msg = (
                "It looks like your SQL statement contains a "
                "potentially dangerous or data-altering operation."
                "If you are positive you want to run this, "
                "pass in acknowledge_risk=True and run this function again."
            )
            raise SQLWarning(msg)
        verbose_message(f"Executing query: {sql}", logger)
        if response == "DICT":
            return self._query_into_dict(sql)
        if response == "DF":
            return self._query_into_df(sql)
        return self._execute_string(sql, ignore_results=(response == "NONE"))

    @abstractmethod
    def get_ddl(self, fqtn: str) -> pd.DataFrame:
        """
        Returns a DataFrame describing the column in the table
        This method should be overridden by derived classes since optimal DDL
        selection will vary by DB type
        `fqtn`: str:
            Fully-qualified Table Name (database.schema.table)
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        db_name, schema_name, table_name = self.parse_fqtn(fqtn)
        sql = (
            f"select table_schema, table_name, column_name, data_type, "
            f"character_maximum_length, column_default, is_nullable from "
            f"INFORMATION_SCHEMA.COLUMNS where table_name = '{table_name}' "
            f"and table_schema = '{schema_name}';"
        )
        query_response = self.execute_query(sql, response="DF")
        return query_response

    @abstractmethod
    def get_object_details(self, fqtn: str) -> tuple:
        """
        Return details of a table or view
        This method should be overridden by derived classes since optimal DDL
        selection will vary by DB type
        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        Response:
            object exists: bool
            is rasgo object: bool
            object type: [table|view|unknown]
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        database, schema, table = self.parse_fqtn(fqtn)
        sql = f"SHOW OBJECTS LIKE '{table}' IN {database}.{schema}"
        result = self.execute_query(sql, response="dict")
        obj_exists = len(result) > 0
        is_rasgo_obj = False
        obj_type = "unknown"
        if obj_exists:
            is_rasgo_obj = result[0].get("comment") == "rasgoql"
            obj_type = result[0].get("kind")
        return obj_exists, is_rasgo_obj, obj_type

    def get_schema(self, fqtn: str, create_sql: str = None) -> List[Tuple[str, str]]:
        """
        Return the schema of a table or view
        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        `create_sql`: str:
            A SQL select statement that will create the view. If this param is passed
            and the fqtn does not already exist, it will be created and profiled based
            on this statement. The view will be dropped after profiling
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        database, schema, table = self.parse_fqtn(fqtn)
        query_sql = (
            f"SELECT * FROM INFORMATION_SCHEMA.TABLES "
            f"WHERE table_catalog = '{database}' "
            f"AND table_schema = '{schema}' "
            f"AND table_name = '{table}'"
        )
        response = []
        try:
            if self._table_exists(fqtn):
                query_response = self.execute_query(query_sql, response="dict")
            elif create_sql:
                self.create(create_sql, fqtn, table_type="view")
                query_response = self.execute_query(query_sql, response="dict")
                self.execute_query(
                    f"DROP VIEW {schema}.{table}",
                    response="none",
                    acknowledge_risk=True,
                )
            else:
                raise TableAccessError(f"Table {fqtn} does not exist or cannot be accessed.")
            for row in query_response:
                response.append((row["name"], row["type"]))
            return response
        except Exception as e:
            self._error_handler(e)

    def list_tables(self, database: str = None, schema: str = None) -> pd.DataFrame:
        """
        List all tables and views available in default namespace
        Params:
        `database`: str:
            override database
        `schema`: str:
            override schema
        """
        select_clause = (
            "SELECT TABLE_NAME, "
            "TABLE_CATALOG||'.'||TABLE_SCHEMA||'.'||TABLE_NAME AS FQTN, "
            "CASE TABLE_TYPE WHEN 'BASE TABLE' THEN 'TABLE' ELSE TABLE_TYPE END AS TABLE_TYPE"
        )
        from_clause = " FROM INFORMATION_SCHEMA.TABLES "
        if database:
            from_clause = f" FROM {database.upper()}.INFORMATION_SCHEMA.TABLES "
        where_clause = f" WHERE TABLE_SCHEMA = '{schema.upper()}'" if schema else ""
        sql = select_clause + from_clause + where_clause
        return self.execute_query(sql, response="df", acknowledge_risk=True)

    def preview(self, sql: str, limit: int = 10) -> pd.DataFrame:
        """
        Returns 10 records into a pandas DataFrame
        Params:
        `sql`: str:
            SQL statment to run
        `limit`: int:
            Records to return
        """
        return self.execute_query(f"{sql} LIMIT {limit}", response="df", acknowledge_risk=True)

    def save_df(self, df: pd.DataFrame, fqtn: str, method: str = None) -> str:
        """
        Creates a table from a pandas Dataframe
        Params:
        `df`: pandas DataFrame:
            DataFrame to upload
        `fqtn`: str:
            Fully-qualied table name (database.schema.table)
            Name for the new table
        `method`: str
            Values: [append, replace]
            when this table already exists in your DataWarehouse,
            pass append: to add dataframe rows to it
            pass replace: to overwrite it with dataframe rows
                WARNING: This will completely overwrite data in the existing table
        """
        if method:
            method = check_write_method(method)
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        database, schema, table = self.parse_fqtn(fqtn)
        table_exists = self._table_exists(fqtn)
        if table_exists and not method:
            msg = (
                f"A table named {fqtn} already exists. "
                "If you are sure you want to write over it, pass in "
                "method='append' or method='replace' and run this function again"
            )
            raise TableConflictException(msg)
        try:
            cleanse_sql_dataframe(df)
            # If the table does not exist or we've received instruction to replace
            # Issue a create or replace statement before we insert data
            if not table_exists or method == "REPLACE":
                create_stmt = generate_dataframe_ddl(df, fqtn)
                self.execute_query(create_stmt, response="None", acknowledge_risk=True)
            df.to_sql(
                table,
                self._engine,
                schema=schema,
                if_exists=method.lower(),
                index=False,
                chunksize=1000,
            )
            return fqtn
        except Exception as e:
            self._error_handler(e)

    # ---------------------------
    # Core Data Warehouse helpers
    # ---------------------------
    def _table_exists(self, fqtn: str) -> bool:
        """
        Check for existence of fqtn in the Data Warehouse and return a boolean
        Params:
        `fqtn`: str:
            Fully-qualified table name (database.schema.table)
        """
        fqtn = self.magic_fqtn_handler(fqtn, self.default_namespace)
        do_i_exist, _, _ = self.get_object_details(fqtn)
        return do_i_exist

    # --------------------------
    # SQLAlchemy and derived class helpers
    # --------------------------
    @property
    @abstractmethod
    def _engine(self):
        """
        Returns a SQLAlchemy engine
        """
        engine_url = (
            f"{self.credentials.get('dw_type')}://"
            f"{self.credentials.get('username')}:"
            f"{urlquote(self.credentials.get('password'))}"
            f"@{self.credentials.get('host')}:"
            f"{self.credentials.get('port')}/"
            f"{self.credentials.get('database')}"
        )
        return alchemy_engine(engine_url)

    def _error_handler(self, exception: Exception, query: str = None) -> None:
        """
        Handle SQLAlchemy exceptions that need additional info
        """
        verbose_message(f"Exception occurred while running query: {query}", logger)
        if exception is None:
            return
        if isinstance(exception, alchemy_exceptions.DisconnectionError):
            raise DWConnectionError(
                "Disconnected from DataWarehouse. Please validate connection or reconnect."
            ) from exception
        raise exception

    def _execute_string(self, query: str, ignore_results: bool = False) -> list[tuple]:
        """
        Execute a query string against the DataWarehouse connection and fetch all results
        """
        try:
            results = self.connection.execute(query)
            if ignore_results:
                return
            return list(results)
        except Exception as e:
            self._error_handler(e)

    def _query_into_dict(self, query: str) -> list[dict]:
        """
        Run a query string and return results in a Snowflake DictCursor
        PRO:
        Results are callable by column name
        for row in data:
            row['COL_NAME']
        CON:
        Query string must be a single statement (only one ;) or Snowflake returns an error
        """
        try:
            query_return = self.connection.execute(query).__dict__
            return query_return
        except Exception as e:
            self._error_handler(e)

    def _query_into_df(self, query: str) -> pd.DataFrame:
        """
        Run
        """
        try:
            query_result = self.connection.execute(query)
            query_return_df = pd.DataFrame(query_result.all())
            response_cols = list(query_result.keys())
            query_return_df.columns = response_cols
            return query_return_df
        except Exception as e:
            self._error_handler(e)
